import json
from typing import Dict, List, Union

from workspaces.workflow.schema import (
    Workflow,
    WorkflowRequest,
    WorkflowRequestFile,
    WorkflowTemplate,
)
from workspaces.workflow.schema_interfaces import (
    WorkflowIF,
    WorkflowRequestFileIF,
    WorkflowRequestIF,
)
from workspaces.workflow.services.interfaces import WorkflowInfoIF


class FakeWorkflowInfo(WorkflowInfoIF):
    """
    A fake workflow info. Providing this as the substrate to other workflow
    services for testing will enable us to ensure that the right things are
    happening under the hood.
    """

    def __init__(self):
        self.requests = []

    def lookup_workflow_request(self, request_id: int) -> WorkflowIF:
        return self.requests[request_id - 1]

    def all_workflows(self) -> List[WorkflowIF]:
        return [self.lookup_workflow_definition("null")]

    def create_workflow_request(
        self, workflow: Union[str, WorkflowIF], argument: Dict
    ) -> WorkflowRequestIF:
        workflow_name = (
            workflow.workflow_name if hasattr(workflow, "workflow_name") else workflow
        )
        request = WorkflowRequest(workflow_name=workflow_name, argument=argument)
        self.save_request(request)
        return request

    def save_request(self, request: WorkflowRequestIF):
        request.workflow_request_id = len(self.requests) + 1
        self.requests.append(request)

    def save_file(
        self, request: WorkflowRequestIF, filename: str, content: bytes
    ) -> WorkflowRequestFileIF:
        request.files.append(WorkflowRequestFile(filename=filename, content=content))

    def lookup_workflow_definition(self, name: str) -> "Workflow":
        if name == "null":
            w = Workflow(workflow_name="null")
            w.templates.append(
                WorkflowTemplate(
                    filename="null.condor",
                    content=b'executable = null\narguments = "-n"\nlog = condor.log\n\nqueue',
                )
            )
            return w
        else:
            return NotImplementedError


info = FakeWorkflowInfo()


def test_workflow_creation():
    # FIXME: This is a unit test, but it seems to only be testing this mocked FakeWorkflowInfo object and not be testing
    #  the actual system in any meaningful way
    # This seems like a bit of a blunt instrument for testing the service. We should probably break the service down
    # into smaller pieces.
    global info

    # create request
    request = info.create_workflow_request(
        info.lookup_workflow_definition("null"), json.loads('{"argument":"-n"}')
    )

    # did the workflow info receive a new workflow request?
    assert request in info.requests
